"use strict";

var _interopRequireWildcard = require("@babel/runtime-corejs3/helpers/interopRequireWildcard");

var _interopRequireDefault = require("@babel/runtime-corejs3/helpers/interopRequireDefault");

var _Object$defineProperty = require("@babel/runtime-corejs3/core-js-stable/object/define-property");

_Object$defineProperty(exports, "__esModule", {
  value: true
});

exports.ContractEnhanced = void 0;

var _find = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/find"));

var _keys = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/object/keys"));

var _sort = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/sort"));

var _map = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/map"));

var _parseInt2 = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/parse-int"));

var _filter = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/filter"));

var _bind = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/bind"));

var _forEach = _interopRequireDefault(require("@babel/runtime-corejs3/core-js-stable/instance/for-each"));

var fs = _interopRequireWildcard(require("fs-extra"));

var path = _interopRequireWildcard(require("path"));

var _solidityParserAntlr = _interopRequireDefault(require("solidity-parser-antlr"));

var _eventId = require("./eventId");

var _injector = require("./injector");

var _instrumenter = require("./instrumenter");

var _instrumentWalker = require("./instrumentWalker");

var _path2 = require("./path");

var _printer = require("./printer");

const STATEMENT_EVENT = "__StatementCoverage";
const POINT_FACTOR = 1000000000;
let id = 0;

function nextId() {
  id++;
  return id;
}

class ContractEnhanced {
  constructor(filepath, solcVersion) {
    this.filepath = filepath;
    this.solcVersion = solcVersion;
    this.id = void 0;
    this.coverage = void 0;
    this.coverageFilepath = void 0;
    this.originalSource = void 0;
    this.source = void 0;
    this.ast = void 0;
    this.functionsBodyLocation = {};
    this.id = nextId(); // silence compiler warnings.

    this.source = this.originalSource = "";

    try {
      this.source = fs.readFileSync(filepath, "utf-8");
      this.originalSource = this.source;
      this.ast = _solidityParserAntlr.default.parse(this.source, {
        loc: true,
        range: true
      });
    } catch (error) {
      const {
        line,
        column,
        message
      } = error.errors[0];
      console.warn(`Error on ${this.filepath}:${line}:${column}: "${message}". Could not setup for coverage.`);
    }

    this.coverageFilepath = path.join((0, _path2.coverageContractsPath)(), this.filepath);
    this.coverage = {
      b: {},
      branchMap: {},
      code: this.originalSource,
      f: {},
      fnMap: {},
      l: {},
      path: filepath,
      s: {},
      statementMap: {}
    };
  }

  instrument() {
    var _context, _context2;

    if (!this.ast) {
      return;
    }

    const instrumenter = new _instrumenter.Instrumenter(this);
    const instrumentWalker = new _instrumentWalker.InstrumentWalker(instrumenter);
    instrumentWalker.walk(this.ast);
    const injector = new _injector.Injector(this, this.solcVersion);
    (0, _forEach.default)(_context = instrumenter.getInjectionPoints()).call(_context, (0, _bind.default)(_context2 = injector.process).call(_context2, injector));
  }

  save() {
    fs.ensureFileSync(this.coverageFilepath);

    if (!this.ast) {
      return fs.writeFileSync(this.coverageFilepath, this.source);
    }

    try {
      const printer = new _printer.Printer(this.ast);
      this.source = printer.print();
    } catch (err) {
      // Something might have happened with the printer. Write the original source.
      console.warn(`Error with coverage printer for ${this.filepath}. Please submit an issue with this code.`);
    } finally {
      fs.writeFileSync(this.coverageFilepath, this.source);
    }
  }

  updateCoverage(events) {
    var _context3;

    (0, _forEach.default)(_context3 = (0, _filter.default)(events).call(events, this.filterCoverageEvent)).call(_context3, event => {
      const value = (0, _parseInt2.default)(event.returnValues[0], 10);
      const {
        contractId,
        injectionPointId
      } = (0, _eventId.decrypt)(value);

      if (contractId !== this.id) {
        return;
      }

      this.coverage.s[injectionPointId] += 1;
      const location = this.coverage.statementMap[injectionPointId];
      this.coverage.l[location.start.line] += 1;
      const fnMapId = this.findFnMapId(location);

      if (fnMapId) {
        this.coverage.f[fnMapId] += 1;
      }

      const [fnBranchId, locationId] = this.findFnBranchIdAndLocationId(location);

      if (fnBranchId) {
        this.coverage.b[fnBranchId][locationId] += 1;
      }
    });
  }

  addStatement(location) {
    const coverageId = this.getNewCoverageId(this.coverage.statementMap);
    this.coverage.statementMap[coverageId] = location;
    this.coverage.s[coverageId] = 0;
    this.coverage.l[location.start.line] = 0;
    return coverageId;
  }

  addBranch(line, type, locations) {
    const coverageId = this.getNewCoverageId(this.coverage.branchMap);
    this.coverage.branchMap[coverageId] = {
      line,
      locations,
      type
    };
    this.coverage.b[coverageId] = (0, _map.default)(locations).call(locations, () => 0);
    this.coverage.l[line] = 0;
    return coverageId;
  }

  addFunction(location, name, bodyLocation) {
    const coverageId = this.getNewCoverageId(this.coverage.fnMap);
    const line = location.start.line;
    this.coverage.fnMap[coverageId] = {
      decl: location,
      line,
      loc: location,
      name
    };
    this.coverage.f[coverageId] = 0;
    this.functionsBodyLocation[coverageId] = bodyLocation;
    return coverageId;
  }

  getNewCoverageId(object) {
    var _context4, _context5;

    const lastId = (0, _sort.default)(_context4 = (0, _map.default)(_context5 = (0, _keys.default)(object)).call(_context5, Number)).call(_context4, (a, b) => b - a)[0] || 0;
    return lastId + 1;
  }

  filterCoverageEvent(event) {
    return STATEMENT_EVENT === event.event;
  }

  findFnMapId(location) {
    var _context6;

    const result = (0, _find.default)(_context6 = (0, _keys.default)(this.functionsBodyLocation)).call(_context6, value => {
      const bodyLocation = this.functionsBodyLocation[(0, _parseInt2.default)(value, 10)];
      return this.isWithin(location.start, bodyLocation);
    });

    if (result) {
      return (0, _parseInt2.default)(result, 10);
    }

    return 0;
  }

  findFnBranchIdAndLocationId(location) {
    var _context7;

    let fnBranchId = 0;
    let locationId = 0;
    (0, _forEach.default)(_context7 = (0, _keys.default)(this.coverage.branchMap)).call(_context7, value => {
      var _context8;

      if (fnBranchId && locationId) {
        return;
      }

      const branch = this.coverage.branchMap[(0, _parseInt2.default)(value, 10)];
      (0, _forEach.default)(_context8 = branch.locations).call(_context8, (branchLocation, index) => {
        if (this.isWithin(location.start, branchLocation)) {
          fnBranchId = (0, _parseInt2.default)(value, 10);
          locationId = index;
        }
      });
    });
    return [fnBranchId, locationId];
  }

  isWithin(point, location) {
    const toCompare = this.getComparablePoint(point);
    return toCompare >= this.getComparablePoint(location.start) && toCompare <= this.getComparablePoint(location.end);
  }

  getComparablePoint(point) {
    return point.line * POINT_FACTOR + point.column;
  }

}

exports.ContractEnhanced = ContractEnhanced;
//# sourceMappingURL=contractEnhanced.js.map